/***********************************************************************

  webtty.js - client side script used to display output from server and 
              to send keypresses to server

  This file is part of the webtty package. The webtty package was
  written by Martin Steen Nielsen

  Development is hosted by testape.com at www.testape.com/webtty

  webtty is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  webtty is distributed in the hope that it will be useful,but 
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with webtty; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

***********************************************************************/
var webTtyTaskArray= new Array;


function fullURIEncode(key)
{
  // return encodeURICOmponent(key);
  return "%"+("00" + key.toString(16)).match(/([0-9A-Fa-f][0-9A-Fa-f])$/)[0];
}


/***********************************************************************
   closes terminal
***********************************************************************/
function wclose(id)
{
  var elem = document.getElementById(id);
  elem.style.display='none';
  return false;
}




/***********************************************************************
   constructor webtty is called from client. Constructor expects to 
   receive id of element that will hold the html for the new process. 
   Function exit_func(err) is optional, but if given, it will be called 
   on errors or if the server terminates the process early.

   Sample usage -->

   <html>
     <script type="text/javascript" src="webtty.js">
     <body onLoad="new webtty('webtty_element');">
       <div id="webtty_element"></div>
     </body>
   </html>


***********************************************************************/
function zeroOrText(s){
  if (s == "\0")
            return "";
  else
    return s;
};

function webtty(receiving_elementID, exit_func)
{
  /********
           Escape symbol control
  ********/
  this.ctrlKey = false
    this.receiving_elementID = receiving_elementID;
    /***********************************************************************
       Creates XMLHTTPRequest object
    ***********************************************************************/
    this.create_transfer_object = function()
    {
      try { return new ActiveXObject('Msxml2.XMLHTTP');    } catch(e) {}
      try { return new ActiveXObject('Microsoft.XMLHTTP'); } catch(e) {}
      try { return new XMLHttpRequest();                   } catch(e) {}
      return null;
    }

  /***********************************************************************
       callback function whenever html data is ready
  ***********************************************************************/
  this.get_html_state = function()
    {
      if (this.xhr_out.readyState == 4)
        {
          //alert("status:"+this.xhr_out.status);
          if ((this.xhr_out.status!=200)
              &&(this.xhr_out.status!=204)
              &&(this.xhr_out.status!=0))
            this.error();
          if (this.xhr_out.status==200 || this.xhr_out.status==204) {
            this.initCounter = this.initCounter + 1;
            //alert('responseText:'+zeroOrText(this.xhr_out.responseText));
            hiddenDiv.innerHTML = zeroOrText(this.xhr_out.responseText);
            //alert('after\10'+hiddenDiv.innerHTML)
            var response = htmlDecode(getText(hiddenDiv));
            if(this.initCounter==1) {
              this.element.innerHTML = "<textarea id=webtty_output rows=25 cols=80 readonly"+" ondblclick=\"unHide(this);\""+">"+"</textarea>";
              this.session_id  = ":)";
              this.output_area = findChildByTag(this.element, 'textarea',3, "", "");
              this.output_area.id = 'default';
              if (response > ''){
                        this.output_area.value = response;
                        this.output_area.scrollTop=this.output_area.scrollHeight;
              }
              //              alert(this.output_area.value);
              var obj = this;
              this.output_area.onkeypress = function(e) { if (obj.running) return obj.key_event(e); }
              this.output_area.onkeydown  = function(e) { if (obj.running) return obj.special_key_event(e); }

              this.timerID=setTimeout("if(webTtyTaskArray['"+this.receiving_elementID+"'] != null) webTtyTaskArray['"+this.receiving_elementID+"'].get_data();", 200);
            }
            else {
              if(response!='') handle_data(response);
            }
          }
        }
    }

  /***********************************************************************
       Request a new process and HTML data from server
  ***********************************************************************/
  this.get_html = function()
    {
      var obj = this;
      var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
      //      if (null == this.xhr_out)
      this.xhr_out     = this.create_transfer_object();
      this.xhr_out.onreadystatechange = function() {};
      this.xhr_out.open('POST', PAGENAME, true);
      this.xhr_out.onreadystatechange = function() { if (obj.running) obj.get_html_state(); }
      this.xhr_out.setRequestHeader("Content-Type", "application/x-www-form-urlencoded")
      this.xhr_out.send('command=tty&path={'+path+'}&'+AUTH+'&value=&');
    }




  /***********************************************************************
       callback function whenever process data is ready
  ***********************************************************************/
  this.get_data_state = function()
    {
      if (this.xhr_out.readyState == 4)
        {
          if ((this.xhr_out.status!=200)&&(this.xhr_out.status!=204)&&(this.xhr_out.status!=0))
            this.error();
          //alert( "rcv:"+zeroOrText(this.xhr_out.responseText));
          hiddenDiv.innerHTML = zeroOrText(this.xhr_out.responseText);
          //alert("html:"+hiddenDiv.innerHTML );
          var data = htmlDecode(getText(hiddenDiv));
          if (data == '_C_L_O_S_E_D_')
            {
              this.output_area.value = this.output_area.value +  data;
              this.output_area.scrollTop=this.output_area.scrollHeight;
              var rows = this.output_area.value.split("\n").length;
              this.output_area.rows = (rows > 5 ? (rows < 25 ? rows : 25) : 5);
              this.session_id=null;
              if (undefined != this.webtty_exit)
                this.webtty_exit(0);
              this.cleanup();
            }
          else 
            {
              if (data > '')
                this.handle_data(data);
              this.timerID=setTimeout("webTtyTaskArray['"+this.receiving_elementID+"'].get_data()", this.timerDelay);
              this.timerDelay = 1000;
              //this.get_data();
            }
        }
    }


  /***********************************************************************
       Request new data from the server
  ***********************************************************************/
  this.get_data = function()
    {
      if (this.timerID != null)
        clearTimeout( this.timerID);
      this.timerID = null;
      if (null == this.xhr_out)
        this.xhr_out     = this.create_transfer_object();
      var obj = this;
      var out;
      var k;
      var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
      this.xhr_out.onreadystatechange = function() { };
      this.xhr_out.open('POST', PAGENAME, true);
      this.xhr_out.onreadystatechange = function() { return obj.get_data_state(); };
      this.xhr_out.setRequestHeader("Content-Type","application/x-www-form-urlencoded; charset=utf-8");

      k = this.kbd_buffer.search(/[%][%][%]/);
      if(k == 0) {
        out = "";
        this.kbd_buffer = this.kbd_buffer.substr(3);
      }else if(k>0) {
        out = "&value="+this.kbd_buffer.substr(0,k);
        this.kbd_buffer = this.kbd_buffer.substr(k);
      } else {
        out = "&value="+this.kbd_buffer;
      };
      //alert("send request:"+'command=tty&path={'+path+'}&'+AUTH + out + '&');
      this.xhr_out.send('command=tty&path={'+path+'}&'+AUTH + out + '&');
      this.kbd_buffer = '';
    }


  /***********************************************************************
       exit page. Remove handles and terminate session
  ***********************************************************************/
  this.cleanup = function()
    {
      this.running     = false;
      if (this.timerID)
        clearTimeout( this.timerID);
      this.timerID = null;
      webTtyTaskArray[this.receiving_elementID] = null;
      if (this.xhr_kbd)
        {
          this.xhr_kbd.onreadystatechange = function() {};
          this.xhr_kbd.abort();
          delete this.xhr_kbd;
        }
      if (this.xhr_out)
        {
          this.xhr_out.onreadystatechange = function() {};
          this.xhr_out.abort();
          delete xhr_out;
        }
      if (null != this.session_id)
        {
          var exit_req = this.create_transfer_object();
          var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
          exit_req.open('POST', PAGENAME, true);
          exit_req.setRequestHeader("Content-Type","application/x-www-form-urlencoded; charset=utf-8");
          exit_req.send('command=tty&path={'+path+'}&'+AUTH);
          delete exit_req;
          this.session_id=null;
        }
    }
    
  /***********************************************************************
       handles process data
  ***********************************************************************/
  this.handle_data = function(data)
    {
      var txt =  this.handle_special_chars(this.output_area.value + data, 2*data.length).split("\n");
      var i = 0;
      while ( i < txt.length ) {
        if(txt[i].length > 80) {
          txt[i] = txt[i].replace(/(.{80})/g,"$1\n").replace(/[\n]$/, "");
        }
        i = i + 1;
      }
      
      this.output_area.value = txt.join("\n");
      this.output_area.scrollTop=this.output_area.scrollHeight;
      
    }

  /***********************************************************************
       Handle special characters. Simulates applying backspace, bell etc to 
       a string
  ***********************************************************************/
  this.handle_special_chars = function(streng, max_look_back)
    {
      var startpos=0;
      var write_pos=0; 
      var new_str='';
      var subs='';

      if (max_look_back>80)
        max_look_back=80;
        
      if (streng.length>max_look_back)
        startpos = streng.length-max_look_back;

      for (tl=startpos; tl<streng.length; tl++)
        {
          if (streng.charCodeAt(tl)==8)
            { write_pos--; subs = subs.substr(0,write_pos); }
          else if (streng.charCodeAt(tl)==7)
            { ; }
          else
            { write_pos++; subs += streng.charAt(tl); }
        }
      return streng.substr(0,startpos)+subs;
    }


  /***********************************************************************
      buffer data and send to server
  ***********************************************************************/
  this.buffer_next_data = function(cmd)
    {
      this.kbd_buffer += cmd;
      if (this.timerID != null) {
        clearTimeout( this.timerID);
        this.timerID = setTimeout("webTtyTaskArray['"+this.receiving_elementID+"'].get_data()", 10);
      } else {
        this.timerDelay = 10;
      }
    }



  /***********************************************************************
       handles normal keypresses
  ***********************************************************************/
  this.key_event = function(e)
    {
      if (!e) var e = window.event;
      if (e.keyCode) key = e.keyCode;
      else if (e.which) key = e.which;
      if (!(window.event || e.charCode))
        return true;
      /*handle escape symbols for Ctrl X*/
      if ((String.fromCharCode(key) == "^") && !this.ctrlKey) {
        this.ctrlKey = true;
        return false;
      }
      if (this.ctrlKey) {
        this.ctrlKey = false;
        if ( (key > 64) && (key < 96) && (key != 94) ) 
          key = key - 64;
      }
      
      if (key==13 || key == 10) {
          return false;
        }

      this.buffer_next_data(fullURIEncode(key));
      return false; 
    }


  /***********************************************************************
       handles special keypresses
  ***********************************************************************/
  this.special_key_event = function (e)
    {
      if (!e) var e = window.event;
      if (e.keyCode) key = e.keyCode;
      else if (e.which) key = e.which;

      if (key==9)
        { 
          this.buffer_next_data(escape("\t")); 
          return false; 
        }
      else if (key==8)
        { 
          this.buffer_next_data(escape("\010")); 
          return false; 
        }
      else if (key == 40)
        { 
          this.buffer_next_data(escape("\033[B")); 
          return false; 
        }
      else if (key == 38)
        { 
          this.buffer_next_data(escape("\033[A")); 
          return false; 
        }
      else if (key == 38)
        { 
          this.buffer_next_data(escape("\033[A")); 
          return false; 
        }
      else if (key == 13 || key == 10)
        {
          this.buffer_next_data(escape("\r"));
          this.cr = true;
          return false; 
        }
    }


  /***********************************************************************
       handles errors
  ***********************************************************************/
  this.error = function()
    {
      if (undefined != this.webtty_exit)
        this.webtty_exit(1);
      if (this.element)
        this.element.innerHTML="Error";
      this.cleanup();
    }


  this.element =  document.getElementById(receiving_elementID);
  if (null == this.element)
    return this.error();

  this.running     = true;
  this.session_id  = null;
  this.output_area = null;
  this.kbd_buffer  = '';
  this.webtty_exit = exit_func;
  this.timerID = null;
  this.timerDelay = 1000;
  this.initCounter = 0;
  webTtyTaskArray[receiving_elementID] = this;
}


function init_webtty_object(objID, exit_handler) {
  /*if (undefined != exitFunc) 
    exit_handler = exitFunc;
    else
    exit_handler =  function webtty_exit()
    {
    var div_element = obj;
    div_element.innerHTML = 'Refresh page to get a new virtual machine';
    };*/
  a = new webtty(objID, exit_handler);
  a.get_html();
  return a;
}
